/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Win32 API upper level graphics drawing routines
 */
#include "windows.h"
#include "wintern.h"
#include <stdlib.h>

#define MAXSYSCOLORS	29	/* # of COLOR_* system colors*/
#define MAXSTOCKOBJECTS	18	/* # of stock objects*/

static HDC	cliphdc;	/* current window cliprects*/

/* default bitmap for new DCs*/
static BITMAPOBJ default_bitmap = {
	{OBJ_BITMAP, TRUE}, 1, 1, 1, 1, 1, 1
};

HDC WINAPI
GetDCEx(HWND hwnd,HRGN hrgnClip,DWORD flags)
{
	HDC	hdc;

	/* add caching?*/
	hdc = ItemNew(struct hdc);
	if(!hdc)
		return NULL;

	if(!hwnd)		/* handle NULL hwnd => desktop*/
		hwnd = rootwp;

	hdc->psd = &scrdev;
	hdc->hwnd = hwnd;
	if(flags & DCX_DEFAULTCLIP) {
		flags &= ~DCX_DEFAULTCLIP;
		if(hwnd->style & WS_CLIPSIBLINGS)
			flags |= DCX_CLIPSIBLINGS;
		if(hwnd->style & WS_CLIPCHILDREN)
			flags |= DCX_CLIPCHILDREN;
	}
	hdc->flags = flags;
	hdc->bkmode = OPAQUE;
	hdc->bkcolor = RGB(255, 255, 255);	/* WHITE*/
	hdc->textcolor = RGB(0, 0, 0);		/* BLACK*/
	hdc->brush = (BRUSHOBJ *)GetStockObject(WHITE_BRUSH);
	hdc->pen = (PENOBJ *)GetStockObject(BLACK_PEN);
	hdc->font = (FONTOBJ *)GetStockObject(SYSTEM_FONT);
	hdc->bitmap = &default_bitmap;
	hdc->drawmode = R2_COPYPEN;
	hdc->pt.x = 0;
	hdc->pt.y = 0;

	return hdc;
}

HDC WINAPI
GetDC(HWND hwnd)
{
	return GetDCEx(hwnd, NULL, DCX_DEFAULTCLIP);
}

HDC WINAPI
GetWindowDC(HWND hwnd)
{
	return GetDCEx(hwnd, NULL, DCX_WINDOW|DCX_DEFAULTCLIP);
}

/* free a DC allocated from GetDC*/
int WINAPI 
ReleaseDC(HWND hwnd, HDC hdc)
{
	/* don't delete a memory dc on release*/
	if(!hdc || (hdc->psd->flags&PSF_MEMORY))
		return 0;

	if(hdc == cliphdc)
		cliphdc = NULL;
	DeleteObject((HBRUSH)hdc->brush);
	DeleteObject((HPEN)hdc->pen);
	DeleteObject((HFONT)hdc->font);
	/*
	 * We can only select a bitmap in a memory DC,
	 * so bitmaps aren't released except through DeleteDC.
	 */
	DeleteObject((HBITMAP)hdc->bitmap);
	ItemFree(hdc);
	return 1;
}

/* free a dc allocated from CreateCompatibleDC*/
BOOL WINAPI
DeleteDC(HDC hdc)
{
	/* don't delete a normal dc, only memory dc's*/
	if(!hdc || (hdc->psd->flags&PSF_SCREEN))
		return 0;

	/* free allocated memory screen device*/
	free(hdc->psd);

	/* make it look like a GetDC dc, and free it*/
	hdc->psd = &scrdev;
	return ReleaseDC(NULL, hdc);
}

HDC WINAPI 
BeginPaint(HWND hwnd, LPPAINTSTRUCT lpPaint)
{
	HDC	hdc;

	/* first update non-client area*/
	SendMessage(hwnd, WM_NCPAINT, 0, 0L);

#if ERASEMOVE
	/* 
	 * Don't allow windows to repaint while user is moving
	 * a window.  Instead, just erase backgrounds
	 * and indicate delayed painting required, which
	 * will occur after user completes window move.
	 */
	if(dragwp) {
		hdc = NULL;
		lpPaint->fErase = !DefWindowProc(hwnd, WM_ERASEBKGND, 0, 0L);
		hwnd->gotPaintMsg = PAINT_DELAYPAINT;
	} else {
#endif
		hdc = GetDC(hwnd);

		/* erase client background*/
		lpPaint->fErase = !SendMessage(hwnd, WM_ERASEBKGND, (WPARAM)hdc,
			0L);
#if ERASEMOVE
	}
#endif
	lpPaint->hdc = hdc;

	/* update region is full client area for now*/
	GetClientRect(hwnd, &lpPaint->rcPaint);
	return hdc;
}

BOOL WINAPI 
EndPaint(HWND hwnd, CONST PAINTSTRUCT *lpPaint)
{
	ReleaseDC(hwnd, lpPaint->hdc);
#if UPDATEREGIONS
#if ERASEMOVE
	/* don't clear update region until done dragging*/
	if(!dragwp)
#endif
		hwnd->nRcUpdate = 0;
#endif
	return TRUE;
}

COLORREF WINAPI
SetTextColor(HDC hdc, COLORREF crColor)
{
	COLORREF	oldtextcolor;

	if (!hdc)
		return crColor;
	oldtextcolor = hdc->textcolor;
	hdc->textcolor = (COLORVAL)crColor;
	return oldtextcolor;
}

COLORREF WINAPI
SetBkColor(HDC hdc, COLORREF crColor)
{
	COLORREF	oldbkcolor;

	if (!hdc)
		return crColor;
	oldbkcolor = hdc->bkcolor;
	hdc->bkcolor = crColor;
	return oldbkcolor;
}

int WINAPI
SetBkMode(HDC hdc, int iBkMode)
{
	int	oldbkmode;

	if(!hdc)
		return iBkMode;
	oldbkmode = hdc->bkmode;
	hdc->bkmode = iBkMode;
	return oldbkmode;
}

/* FIXME: releasing a DC does NOT change back the drawing mode!*/
int WINAPI
SetROP2(HDC hdc, int fnDrawMode)
{
	int	newmode, oldmode;

	if(!hdc)
		return 0;

	oldmode = hdc->drawmode;
	switch(fnDrawMode) {
	case R2_XORPEN:
		newmode = MODE_XOR;
		break;
	case R2_COPYPEN:
	default:
		newmode = MODE_SET;
	}
	hdc->drawmode = newmode;
	GdSetMode(newmode);
	return oldmode;
}

/* 
 * Setup clip region from device context's associated window or bitmap.
 * Memory DC's are always associated with the desktop window, and are
 * always visible.  Return the DC's hwnd if window is visible.
 */
HWND
WndPrepareDC(HDC hdc)
{
	HWND	hwnd;
	static CLIPRECT crc = {0, 0, 0, 0};

	if(!hdc || !hdc->hwnd)
		return NULL;

	hwnd = hdc->hwnd;
	if (hwnd->unmapcount)
		return NULL;

	/*
	 * If the window is not the currently clipped one, then
	 * make it the current one and define its clip rectangles.
	 */
	if(hdc != cliphdc) {
		/* clip memory dc's to the bitmap size*/
		if(hdc->psd->flags&PSF_MEMORY) {
			crc.width = hdc->psd->xvirtres;
			crc.height = hdc->psd->yvirtres;
			GdSetClipRects(1, &crc);
		} else WndSetClipWindow(hdc);
		cliphdc = hdc;
	}

	return hwnd;
}

COLORREF WINAPI
SetPixel(HDC hdc, int x, int y, COLORREF crColor)
{
	HWND		hwnd;
	POINT		pt;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return 0;	/* doesn't return previous color*/
	pt.x = x;
	pt.y = y;
	if(WndIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	/* draw point in passed color*/
	GdSetForeground(GdFindColor(crColor));
	GdPoint(hdc->psd, pt.x, pt.y);
	return 0;		/* doesn't return previous color*/
}

BOOL WINAPI 
MoveToEx(HDC hdc, int x, int y, LPPOINT lpPoint)
{
	if(!hdc)
		return FALSE;
	if(lpPoint)
		*lpPoint = hdc->pt;
	hdc->pt.x = x;
	hdc->pt.y = y;
	return TRUE;
}

BOOL WINAPI
LineTo(HDC hdc, int x, int y)
{
	HWND		hwnd;
	POINT		beg, end;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	beg.x = hdc->pt.x;
	beg.y = hdc->pt.y;
	end.x = x;
	end.y = y;
	if(WndIsClientDC(hdc)) {
		ClientToScreen(hwnd, &beg);
		ClientToScreen(hwnd, &end);
	}

	/* draw line in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForeground(GdFindColor(hdc->pen->color));
		/* don't draw last point*/
		GdLine(hdc->psd, beg.x, beg.y, end.x, end.y, FALSE);
	}
	hdc->pt.x = x;
	hdc->pt.y = y;
	return TRUE;
}

BOOL WINAPI
Rectangle(HDC hdc, int nLeft, int nTop, int nRight, int nBottom)
{
	HWND	hwnd;
	RECT	rc;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	SetRect(&rc, nLeft, nTop, nRight, nBottom);
	if(WndIsClientDC(hdc)) {
		ClientToScreen(hwnd, (LPPOINT)&rc.left);
		ClientToScreen(hwnd, (LPPOINT)&rc.right);
	}

	/* draw rectangle in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForeground(GdFindColor(hdc->pen->color));
		GdRect(hdc->psd, rc.left, rc.top,
			rc.right - rc.left, rc.bottom - rc.top);
	}

	/* fill rectangle in current brush color*/
	if(hdc->brush->style != BS_NULL) {
		InflateRect(&rc, -1, -1);
		GdSetForeground(GdFindColor(hdc->brush->color));
		GdFillRect(hdc->psd, rc.left, rc.top, rc.right - rc.left,
			rc.bottom - rc.top);
	}

	return TRUE;
}

int WINAPI
FillRect(HDC hdc, CONST RECT *lprc, HBRUSH hbr)
{
	HWND		hwnd;
	RECT 		rc;
	BRUSHOBJ *	obr = (BRUSHOBJ *)hbr;
	COLORREF	crFill;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd || !obr)
		return FALSE;

	if(!lprc) {
		if(WndIsClientDC(hdc))
			GetClientRect(hwnd, &rc);
		else
			GetWindowRect(hwnd, &rc);
		lprc = &rc;
	} else
		rc = *lprc;
	if(WndIsClientDC(hdc)) {
		ClientToScreen(hwnd, (LPPOINT)&rc.left);
		ClientToScreen(hwnd, (LPPOINT)&rc.right);
	}

	/* handle COLOR_xxx + 1 passed as HBRUSH*/
	if((unsigned int)obr <= MAXSYSCOLORS)
		crFill = GetSysColor((int)obr-1);
	else {
		/* get color from passed HBRUSH*/
		if(obr->style == BS_NULL)
			return TRUE;
		crFill = obr->color;
	}

	/* fill rectangle in passed brush color*/
	GdSetForeground(GdFindColor(crFill));
	GdFillRect(hdc->psd, rc.left, rc.top,
		rc.right - rc.left, rc.bottom - rc.top);
	return TRUE;
}

BOOL WINAPI
TextOut(HDC hdc, int x, int y, LPCSTR lpszString, int cbString)
{
	return ExtTextOut(hdc, x, y, 0, NULL, lpszString, cbString, NULL);
}

BOOL WINAPI
ExtTextOut(HDC hdc, int x, int y, UINT fuOptions, CONST RECT *lprc,
	LPCSTR lpszString, UINT cbCount, CONST INT *lpDx)
{
	HWND	hwnd;
	POINT	pt;
	RECT	rc;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	pt.x = x;
	pt.y = y;
	if(WndIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	/* optionally fill passed rectangle*/
	if(lprc && (fuOptions&ETO_OPAQUE)) {
		rc = *lprc;
		if(WndIsClientDC(hdc)) {
			ClientToScreen(hwnd, (LPPOINT)&rc.left);
			ClientToScreen(hwnd, (LPPOINT)&rc.right);
		}
		/* fill rectangle with current background color*/
		GdSetForeground(GdFindColor(hdc->bkcolor));
		GdFillRect(hdc->psd, rc.left, rc.top, rc.right - rc.left,
			rc.bottom - rc.top);
		GdSetUseBackground(FALSE);
	} else {
		/* use current background mode for text background draw*/
		GdSetUseBackground(hdc->bkmode == OPAQUE? TRUE: FALSE);
		if(hdc->bkmode == OPAQUE)
			GdSetBackground(GdFindColor(hdc->bkcolor));
	}

	/* nyi: lpDx*/

	/* draw text in current text foreground and background color*/
	GdSetForeground(GdFindColor(hdc->textcolor));
	GdSetFont(hdc->font->fontno);
	GdText(hdc->psd, pt.x, pt.y, lpszString, cbCount, FALSE);

	return TRUE;
}

/* note: many DT_x aren't implemented in this function*/
int WINAPI
DrawText(HDC hdc, LPCSTR lpString, int nCount, LPRECT lpRect, UINT uFormat)
{
	COORD	width, height;

	if(nCount == -1)
		nCount = strlen(lpString);

	if(uFormat & DT_CALCRECT) {
		if(!hdc)
			return 0;
		GdSetFont(hdc->font->fontno);
		GdGetTextSize(hdc->psd, (char *)lpString, nCount, &width,
			&height);
		lpRect->right = lpRect->left + width;
		lpRect->bottom = lpRect->top + height;
		return height;
	}

	/* draw text at DT_LEFT, DT_TOP using current fg, bg and bkmode*/
	TextOut(hdc, lpRect->left, lpRect->top, lpString, nCount);
	return height;
}

/* Microwindows only*/
BOOL WINAPI
DrawDIB(HDC hdc,int x,int y,PIMAGEHDR pimage)
{
	HWND		hwnd;
	POINT		pt;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd || !pimage)
		return FALSE;
	pt.x = x;
	pt.y = y;
	if(WndIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	GdDrawImage(hdc->psd, pt.x, pt.y, pimage);
	return TRUE;
}

/* define color scheme: A (tan), B (winstd) or C (old)*/
#if ELKS
#define B
#else
#define A
#endif

#define A_RGB(r,g,b)
#define B_RGB(r,g,b)
#define C_RGB(r,g,b)

#ifdef A
#undef  A_RGB
#define A_RGB(r,g,b)	RGB(r,g,b),
#endif
#ifdef B
#undef  B_RGB
#define B_RGB(r,g,b)	RGB(r,g,b),
#endif
#ifdef C
#undef  C_RGB
#define C_RGB(r,g,b)	RGB(r,g,b),
#endif

static COLORREF sysColors[MAXSYSCOLORS] = {
	RGB(192, 192, 192),	/* COLOR_SCROLLBAR          0*/
	RGB(  0, 128, 128),     /* COLOR_BACKGROUND          */
	A_RGB(128,   0,   0)	/* COLOR_ACTIVECAPTION       */
	B_RGB(128,   0, 128)	/* COLOR_ACTIVECAPTION       */
	C_RGB(128,   0, 128)	/* COLOR_ACTIVECAPTION       */
	A_RGB(162, 141, 104)	/* COLOR_INACTIVECAPTION     */
	B_RGB(128, 128, 128)  	/* COLOR_INACTIVECAPTION     */
	C_RGB(  0,  64, 128)	/* COLOR_INACTIVECAPTION     */
	RGB(192, 192, 192),     /* COLOR_MENU                */
	RGB(255, 255, 255),     /* COLOR_WINDOW             5*/
	RGB(  0,   0,   0),     /* COLOR_WINDOWFRAME         */
	RGB(  0,   0,   0),     /* COLOR_MENUTEXT            */
	RGB(  0,   0,   0),     /* COLOR_WINDOWTEXT          */
	RGB(255, 255, 255),     /* COLOR_CAPTIONTEXT         */
	RGB(192, 192, 192),   	/* COLOR_ACTIVEBORDER      10*/
	RGB(192, 192, 192), 	/* COLOR_INACTIVEBORDER      */
	RGB(128, 128, 128),   	/* COLOR_APPWORKSPACE        */
	RGB(128,   0,   0),     /* COLOR_HIGHLIGHT           */
	RGB(255, 255, 255),     /* COLOR_HIGHLIGHTTEXT       */
	A_RGB(213, 204, 187)	/* COLOR_BTNFACE           15*/
	B_RGB(192, 192, 192)	/* COLOR_BTNFACE           15*/
	C_RGB(160, 160, 160)	/* COLOR_BTNFACE           15*/
	A_RGB(162, 141, 104)	/* COLOR_BTNSHADOW           */
	B_RGB(128, 128, 128)	/* COLOR_BTNSHADOW           */
	C_RGB(128, 128, 128)	/* COLOR_BTNSHADOW           */
	RGB( 64,  64,  64),     /* COLOR_GRAYTEXT            */
	RGB(  0,   0,   0),     /* COLOR_BTNTEXT             */
	RGB(192, 192, 192),    	/* COLOR_INACTIVECAPTIONTEXT */
	A_RGB(234, 230, 221)  	/* COLOR_BTNHIGHLIGHT      20*/
	B_RGB(255, 255, 255)  	/* COLOR_BTNHIGHLIGHT      20*/
	C_RGB(223, 223, 223)  	/* COLOR_BTNHIGHLIGHT      20*/
	RGB(  0,   0,   0),     /* COLOR_3DDKSHADOW          */
	A_RGB(213, 204, 187)	/* COLOR_3DLIGHT             */
	B_RGB(223, 223, 223)	/* COLOR_3DLIGHT             */
	C_RGB(192, 192, 192)	/* COLOR_3DLIGHT             */
	RGB(  0,   0,   0),     /* COLOR_INFOTEXT            */
	RGB(225, 255, 255), 	/* COLOR_INFOBK              */
	RGB(184, 180, 184),  	/* COLOR_ALTERNATEBTNFACE  25*/
	RGB(  0,   0, 255),     /* COLOR_HOTLIGHT 	     */
	RGB( 16, 132, 208),  	/* COLOR_GRADIENTACTIVECAPTION */
	RGB(184, 180, 184) 	/* COLOR_GRADIENTINACTIVECAPTION 28*/
};

DWORD WINAPI
GetSysColor(int nIndex)
{
	if(nIndex >= 0 && nIndex < MAXSYSCOLORS)
		return sysColors[nIndex];
	return 0;
}

static BRUSHOBJ OBJ_WHITE_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(255, 255, 255)
};

static BRUSHOBJ OBJ_LTGRAY_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(192, 192, 192)
};

static BRUSHOBJ OBJ_GRAY_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(128, 128, 128)
};

static BRUSHOBJ OBJ_DKGRAY_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(32, 32, 32)
};

static BRUSHOBJ OBJ_BLACK_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(0, 0, 0)
};

static BRUSHOBJ OBJ_NULL_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_NULL, RGB(0, 0, 0)
};

static PENOBJ OBJ_WHITE_PEN = {
	{OBJ_PEN, TRUE}, PS_SOLID, RGB(255, 255, 255)
};

static PENOBJ OBJ_BLACK_PEN = {
	{OBJ_PEN, TRUE}, PS_SOLID, RGB(0, 0, 0)
};

static PENOBJ OBJ_NULL_PEN = {
	{OBJ_PEN, TRUE}, PS_NULL, RGB(0, 0, 0)
};

static FONTOBJ OBJ_OEM_FIXED_FONT = {
	{OBJ_FONT, TRUE}, FONT_OEM_FIXED
};

static FONTOBJ OBJ_ANSI_FIXED_FONT = {
	{OBJ_FONT, TRUE}, FONT_SYSTEM_FIXED
};

static FONTOBJ OBJ_ANSI_VAR_FONT = {
	{OBJ_FONT, TRUE}, FONT_SYSTEM_VAR
};

static FONTOBJ OBJ_SYSTEM_FONT = {
	{OBJ_FONT, TRUE}, FONT_SYSTEM_VAR
};

static FONTOBJ OBJ_DEVICE_DEFAULT_FONT = {
	{OBJ_FONT, TRUE}, FONT_DEVICE_DEFAULT
};

static FONTOBJ OBJ_SYSTEM_FIXED_FONT = {
	{OBJ_FONT, TRUE}, FONT_SYSTEM_FIXED
};

static FONTOBJ OBJ_DEFAULT_GUI_FONT = {
	{OBJ_FONT, TRUE}, FONT_GUI_VAR
};

static struct hgdiobj *stockObjects[MAXSTOCKOBJECTS] = {
	(struct hgdiobj *)&OBJ_WHITE_BRUSH,		/* WHITE_BRUSH*/
	(struct hgdiobj *)&OBJ_LTGRAY_BRUSH,		/* LTGRAY_BRUSH*/
	(struct hgdiobj *)&OBJ_GRAY_BRUSH,		/* GRAY_BRUSH*/
	(struct hgdiobj *)&OBJ_DKGRAY_BRUSH,		/* DKGRAY_BRUSH*/
	(struct hgdiobj *)&OBJ_BLACK_BRUSH,		/* BLACK_BRUSH*/
	(struct hgdiobj *)&OBJ_NULL_BRUSH,		/* NULL_BRUSH*/
	(struct hgdiobj *)&OBJ_WHITE_PEN,		/* WHITE_PEN*/
	(struct hgdiobj *)&OBJ_BLACK_PEN,		/* BLACK_PEN*/
	(struct hgdiobj *)&OBJ_NULL_PEN,		/* NULL_PEN*/
	(struct hgdiobj *)NULL,
	(struct hgdiobj *)&OBJ_OEM_FIXED_FONT,		/* OEM_FIXED_FONT*/
	(struct hgdiobj *)&OBJ_ANSI_FIXED_FONT,		/* ANSI_FIXED_FONT*/
	(struct hgdiobj *)&OBJ_ANSI_VAR_FONT,		/* ANSI_VAR_FONT*/
	(struct hgdiobj *)&OBJ_SYSTEM_FONT,		/* SYSTEM_FONT*/
	(struct hgdiobj *)&OBJ_DEVICE_DEFAULT_FONT,	/* DEVICE_DEFAULT_FONT*/
	(struct hgdiobj *)NULL,				/* DEFAULT_PALETTE*/
	(struct hgdiobj *)&OBJ_SYSTEM_FIXED_FONT,	/* SYSTEM_FIXED_FONT*/
	(struct hgdiobj *)&OBJ_DEFAULT_GUI_FONT		/* DEFAULT_GUI_FONT*/
};

HGDIOBJ WINAPI
GetStockObject(int nObject)
{
	HGDIOBJ	pObj;

	if(nObject >= 0 && nObject < MAXSTOCKOBJECTS) {
		pObj = stockObjects[nObject];

		/* test code for multiple color schemes with
		 * standard background brushes...
		 */
		switch(nObject) {
		case LTGRAY_BRUSH:
		case GRAY_BRUSH:
			((BRUSHOBJ *)pObj)->color = GetSysColor(COLOR_BTNFACE);
			break;
		case DKGRAY_BRUSH:
			((BRUSHOBJ *)pObj)->color = GetSysColor(COLOR_BTNSHADOW);
			break;
		}
		return pObj;
	}
	return NULL;
}

HGDIOBJ WINAPI
SelectObject(HDC hdc, HGDIOBJ hObject)
{
	HGDIOBJ		objOrg;
	BITMAPOBJ *	objBitmap;

	if(!hdc || !hObject)
		return NULL;

	switch(hObject->hdr.type) {
	case OBJ_PEN:
		objOrg = (HGDIOBJ)hdc->pen;
		hdc->pen = (PENOBJ *)hObject;
		break;
	case OBJ_BRUSH:
		objOrg = (HGDIOBJ)hdc->brush;
		hdc->brush = (BRUSHOBJ *)hObject;
		break;
	case OBJ_FONT:
		objOrg = (HGDIOBJ)hdc->font;
		hdc->font = (FONTOBJ *)hObject;
		break;
	case OBJ_BITMAP:
		/* must be memory dc to select bitmap*/
		if(hdc->psd->flags&PSF_SCREEN)
			return NULL;
		objOrg = (HGDIOBJ)hdc->bitmap;
		hdc->bitmap = (BITMAPOBJ *)hObject;

		/* setup mem dc for drawing into bitmap*/
		objBitmap = (BITMAPOBJ *)hObject;

		if (hdc->psd->flags&PSF_PORTRAIT) {
			hdc->psd->yres = objBitmap->width;
			hdc->psd->xres = objBitmap->height;
		} else {
			hdc->psd->xres = objBitmap->width;
			hdc->psd->yres = objBitmap->height;		
		}
		hdc->psd->xvirtres = objBitmap->width;
		hdc->psd->yvirtres = objBitmap->height;
		hdc->psd->planes = objBitmap->planes;
		hdc->psd->bpp = objBitmap->bpp;
		hdc->psd->linelen = objBitmap->linelen;
		hdc->psd->size = objBitmap->size;
		hdc->psd->addr = &objBitmap->bits[0];
	    	break;
	default:
		return NULL;
	}

	return objOrg;
}

BOOL WINAPI
DeleteObject(HGDIOBJ hObject)
{
	if(!hObject || hObject->hdr.stockobj)
		return FALSE;
	/*if(hObject->hdr.type == OBJ_REGION)
		GdDestroyClipRegion(((RGNOBJ *)hObject)->rgn);*/
	ItemFree(hObject);
	return TRUE;
}

HBRUSH WINAPI
CreateSolidBrush(COLORREF crColor)
{
	BRUSHOBJ *hbr;

	hbr = ItemNew(BRUSHOBJ);
	if(!hbr)
		return NULL;
	hbr->hdr.type = OBJ_BRUSH;
	hbr->hdr.stockobj = FALSE;
	hbr->style = BS_SOLID;
	hbr->color = crColor;
	return (HBRUSH)hbr;
}

HPEN WINAPI
CreatePen(int nPenStyle, int nWidth, COLORREF crColor)
{
	PENOBJ *hpen;

	/* fix: nWidth > 1*/
	hpen = ItemNew(PENOBJ);
	if(!hpen)
		return NULL;
	hpen->hdr.type = OBJ_PEN;
	hpen->hdr.stockobj = FALSE;
	hpen->style = nPenStyle;
	hpen->color = crColor;
	return (HPEN)hpen;
}

HBITMAP WINAPI
CreateCompatibleBitmap(HDC hdc, int nWidth, int nHeight)
{
	BITMAPOBJ *	hbitmap;
	int		size;
	int		linelen, bytelen;
	int		tmp;

	/* calc size of memory allocation*/
	nWidth = max(nWidth, 1);
	nHeight = max(nHeight, 1);

	/* 
	 * swap width and height in portrait mode,
	 * so imagesize is calculated properly
	 */
	if(hdc->psd->flags & PSF_PORTRAIT) {
		tmp = nWidth;
		nWidth = nHeight;
		nHeight = tmp;
	}

	/*
	 * Ignore hdc, use screen bpp, planes to create size and linelen.
	 * Linelen is in bytes for bpp 1, 2, 4, 8, and pixels for bpp 16, 32.
	 */
	/* FIXME: move this calc to driver level*/
	bytelen = linelen = nWidth;	/* planes != 1*/
	if(scrdev.planes == 1) {
		switch(scrdev.bpp) {
		case 1:
			bytelen = linelen = (nWidth+7)/8;
			break;
		case 2:
			bytelen = linelen = (nWidth+3)/4;
			break;
		case 4:
			bytelen = linelen = (nWidth+1)/2;
			break;
		case 8:
			bytelen = linelen = nWidth;
			break;
		case 16:
			linelen = nWidth;
			bytelen = nWidth * 2;
			break;
		case 32:
			linelen = nWidth;
			bytelen = nWidth * 4;
			break;
		}
		size = bytelen * nHeight;
	} else {
		/* FIXME always assumes VGA 4 planes 4bpp*/
		/* we use 4bpp linear for memdc format*/
		bytelen = linelen = (nWidth+1)/2;
		size = bytelen * nHeight;
	}

	/* swap width and height back if portrait*/
	if(hdc->psd->flags & PSF_PORTRAIT) {
		tmp = nWidth;
		nWidth = nHeight;
		nHeight = tmp;
	}

	/* allocate gdi object*/
	hbitmap = (BITMAPOBJ *)ItemAlloc(sizeof(BITMAPOBJ)-1+size);
	if(!hbitmap)
		return NULL;
	hbitmap->hdr.type = OBJ_BITMAP;
	hbitmap->hdr.stockobj = FALSE;
	hbitmap->width = nWidth;
	hbitmap->height = nHeight;

	/* ignore hdc for now, create compatible with screen*/
	hbitmap->planes = scrdev.planes;
	hbitmap->bpp = scrdev.bpp;
	hbitmap->linelen = linelen;
	hbitmap->size = size;

	return (HBRUSH)hbitmap;
}

/* return NULL if no driver bitblit available*/
HDC WINAPI
CreateCompatibleDC(HDC hdc)
{
	HDC	hdcmem;
	PSD	psd;

	/* ignore hdc for now, make compatible with screen*/
	/* if driver doesn't have blit, we'll fail*/
	if((scrdev.flags & PSF_HAVEBLIT) == 0)
		return NULL;
	psd = ItemNew(SCREENDEVICE);
	if(!psd)
		return NULL;

	/* allocate a DC for DesktopWindow*/
	hdcmem = GetDCEx(NULL, NULL, DCX_DEFAULTCLIP);
	if(!hdcmem) {
		free(psd);
		return NULL;
	}

	/* copy screen device*/
	*psd = scrdev;

	/* replace with allocated memory driver*/
	psd->flags |= PSF_MEMORY;
	psd->flags &= ~PSF_SCREEN;
	hdcmem->psd = psd;

	/* select in default bitmap to setup mem device parms*/
	SelectObject(hdcmem, (HGDIOBJ)&default_bitmap);
	return hdcmem;
}

BOOL WINAPI
BitBlt(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight,
	HDC hdcSrc, int nXSrc, int nYSrc, DWORD dwRop)
{
	HWND	hwnd;
	POINT	dst, src;

	if(!hdcDest || !hdcSrc)
		return FALSE;
	dst.x = nXDest;
	dst.y = nYDest;
	src.x = nXSrc;
	src.y = nYSrc;

	/* if src screen DC, convert coords*/
	/* FIXME: src clipping isn't checked, only one set of cliprects also*/
	if(!WndIsMemDC(hdcSrc) && WndIsClientDC(hdcSrc)) {
		if(!(hwnd = WndPrepareDC(hdcSrc)))
			return FALSE;
		ClientToScreen(hwnd, &src);
	}
	/* if dst screen DC, convert coords and set clipping*/
	/* FIXME: if dest is also screen, src clipping will be overwritten*/
	if(!WndIsMemDC(hdcDest) && WndIsClientDC(hdcDest)) {
		if(!(hwnd = WndPrepareDC(hdcDest)))
			return FALSE;
		ClientToScreen(hwnd, &dst);
	}

	GdBlit(hdcDest->psd, dst.x, dst.y, nWidth, nHeight,
		hdcSrc->psd, src.x, src.y, dwRop);
	return TRUE;
}

UINT WINAPI
GetSystemPaletteEntries(HDC hdc,UINT iStartIndex,UINT nEntries,
	LPPALETTEENTRY lppe)
{
	UINT		i;
	RGBENTRY	rgb;

	/* currently, we only work for screen device*/
	if(!hdc || hdc->psd != &scrdev)
		return 0;

	for(i=0; i<nEntries; ++i) {
		if(!GdGetPalette(i+iStartIndex, 1, &rgb))
			break;
		lppe->peRed = rgb.r;
		lppe->peGreen = rgb.g;
		lppe->peBlue = rgb.b;
		lppe->peFlags = 0;
		++lppe;
	}
	return i;
}
